/** @file gui_base.h

	Skelton for retropc emulator

	@author Sasaji
	@date   2015.04.01

	@brief [ gui base ]
*/

#ifndef _GUI_BASE_H_
#define _GUI_BASE_H_

#include <windows.h>
#include "../common.h"
#include "../emu.h"
#include "../vm/vm.h"

class EMU;

/// GUI wrapper class
class GUI_BASE
{
protected:
	EMU *emu;
	HWND hWindow;
	int need_update_screen;
	bool globalkey_enable;

public:
	GUI_BASE(int argc, char **argv, EMU *new_emu);
	virtual ~GUI_BASE();

	/// @name send to gui from emu class or main.
	//@{

	/// @return -1:error 0:normal ok 1:ok but use default locale
	virtual int Init();
#ifdef GUI_USE_FOREIGN_WINDOW
	/// create window without SDL library
	virtual int CreateWindow(int width, int height);
#endif
	/// create widget if need
	virtual int CreateWidget(HWND new_hwnd, int width, int height);
	/// create menu if need
	/// @return -1:error 0:normal ok 1:ok (need window resize)
	virtual int CreateMenu();
	/// show gui menu
	virtual void ShowMenu();
	/// hide gui menu
	virtual void HideMenu();

	/// draw gui widget on surface
	/// @return 1:processed SDL_UpdateRect() in this function
	virtual int MixSurface();

	/// post update screen request message
	/// @attention this function is called by emu thread
	/// @return false:did not call UpdateScreen in main thread
	virtual bool NeedUpdateScreen();
	/// post finished updateing screen request message
	/// @attention this function is called by emu thread
	virtual void UpdatedScreen();
	/// draw screen on window
	/// @note this function should be called by main thread
	virtual void UpdateScreen(HDC hdc);

	virtual int CreateGlobalKeys();

	/// @return true : processed global key
	virtual bool ExecuteGlobalKeys(int code, uint32 status);
	/// @return true : processed global key
	virtual bool ReleaseGlobalKeys(int code, uint32 status);

	/// something do before process event
	virtual void PreProcessEvent();
	/// event processing
	/// @return -1: exit event loop 0: processed myself 1: no processing
	virtual LRESULT ProcessEvent(UINT iMsg, WPARAM wParam, LPARAM lParam);
	/// something do after process event
	virtual void PostProcessEvent();

	/// callback from emu screen change
	virtual void ScreenModeChanged(bool fullscreen);

	/// post to WM_COMMAND event message
	void PostCommandMessage(int id, void *data1 = NULL, void *data2 = NULL);

	///
	virtual void KeyDown(int code, uint32 mod);
	///
	virtual void KeyUp(int code);

	///
	virtual void SetFocusToMainWindow();

	//@}

	/// @name send to ui class from ui (user / operator).
	//@{
	virtual void Exit(void);

	// Show Dialog

	virtual bool ShowLoadDataRecDialog(void);
	virtual bool ShowSaveDataRecDialog(void);

	virtual bool ShowOpenFloppyDiskDialog(int drv);

	virtual bool ShowLoadStateDialog(void);
	virtual bool ShowSaveStateDialog(bool cont);

	virtual bool ShowOpenAutoKeyDialog(void);

	virtual bool ShowPlayRecKeyDialog(void);
	virtual bool ShowRecordRecKeyDialog(void);

	virtual bool ShowRecordStateAndRecKeyDialog(void);

	virtual bool ShowSavePrinterDialog(int drv);

	virtual bool ShowRecordVideoDialog(int num);

	virtual bool ShowVolumeDialog(void);
	virtual bool ShowSndFilterDialog(void);

	virtual bool ShowKeybindDialog(void);
	virtual bool ShowConfigureDialog(void);

	virtual bool ShowAboutDialog(void);
	//@}

	/// @name callback to emu class from ui (user / operator).
	//@{

	// Control

	virtual void PostEtReset(void);
	virtual void PostEtSpecialReset(void);
	virtual void PostEtWarmReset(int onoff);
	virtual void Dipswitch(int bit);
	virtual uint8 GetDipswitch(void);
	virtual void PostEtTogglePause(void);
	virtual void TogglePause(void);
	virtual bool NowPause(void);
	virtual void PostEtSystemPause(bool val);
	virtual void SystemPause(bool val);
	virtual bool NowSystemPause(void);
	virtual void PostEtCPUPower(int num);
	virtual int  GetCPUPower(void);

	virtual void ChangeFddType(int num);
	virtual int  GetFddType(void);
	virtual int  NextFddType(void);
	virtual void PostEtToggleSyncIRQ(void);
	virtual bool NowSyncIRQ(void);

	// Screen

	virtual void ChangeFrameRate(int num);
	virtual int  GetFrameRateNum(void);
	virtual void ChangeWindowMode(int num);
	virtual void ChangeFullScreenMode(int num);
	virtual int  GetWindowMode(void);
	virtual int  GetFullScreenMode(void);
	virtual int  GetWindowModeCount(void);
	virtual bool GetWindowModeStr(int num, _TCHAR *str);
	virtual int  GetFullScreenModeCount(void);
	virtual bool GetFullScreenModeStr(int num, _TCHAR *str);
	virtual bool IsFullScreen(void);
	virtual void PostEtToggleScanLine(void);
	virtual bool NowScanLine(void);
	virtual void ToggleStretchScreen(void);
	virtual bool NowStretchedScreen(void);
	virtual void ToggleCutoutScreen(void);
	virtual bool NowCutoutedScreen(void);

	virtual void PostEtStartRecordVideo(int num);
	virtual void PostEtStopRecordVideo(void);
	virtual bool NowRecordingVideo(void);
	virtual int  GetRecordVideoFrameNum(void);
	virtual void PostEtResizeRecordVideoSurface(int num);
	virtual int  GetRecordVideoSurfaceNum(void);
	virtual bool GetRecordVideoSizeStr(int num, _TCHAR *str);
	virtual void PostEtCaptureScreen(void);

	virtual void PostEtChangeAfterImage(int num);
	virtual void PostEtChangeAfterImage(void);
	virtual int  GetAfterImageMode(void);
	virtual void PostEtChangeKeepImage(int num);
	virtual void PostEtChangeKeepImage(void);
	virtual int  GetKeepImageMode(void);

#ifdef USE_DIRECT3D
	virtual void ChangeUseDirect3D(int num);
	virtual void ChangeUseDirect3D(void);
	virtual int  GetDirect3DMode(void);
#endif

	// Sound

	virtual void PostEtStartRecordSound(void);
	virtual void PostEtStopRecordSound(void);
	virtual bool NowRecordingSound(void);
	virtual void ChangeSoundFrequency(int num);
	virtual int  GetSoundFrequencyNum();
	virtual void ChangeSoundLatency(int num);
	virtual int  GetSoundLatencyNum();

#ifdef USE_DATAREC
	// Data Recorder Tape

	virtual void PostEtLoadDataRecMessage(const _TCHAR *file_path);
	virtual void PostEtLoadRecentDataRecMessage(int num);
	virtual void PostEtSaveDataRecMessage(const _TCHAR *file_path);
	virtual void PostEtCloseDataRecMessage(void);
	virtual void PostEtRewindDataRecMessage(void);
	virtual void PostEtFastForwardDataRecMessage(void);
	virtual void PostEtToggleRealModeDataRecMessage(void);
	virtual bool NowRealModeDataRec(void);
	virtual bool IsOpenedLoadDataRecFile(void);
	virtual bool IsOpenedSaveDataRecFile(void);
#endif

#ifdef USE_FD1
	// Floppy disk

	virtual void PostEtOpenFloppyMessage(int drv, const _TCHAR *file_path, int bank, uint32 flags, bool multiopen);
	virtual void PostEtOpenRecentFloppyMessage(int drv, int num);
	virtual void PostEtOpenFloppySelectedVolume(int drv, int bank_num);
	virtual void PostEtCloseFloppyMessage(int drv);
	virtual void PostEtChangeSideFloppyDisk(int drv);
	virtual int  GetSideFloppyDisk(int drv);
	virtual void PostEtToggleWriteProtectFloppyDisk(int drv);
	virtual bool InsertedFloppyDisk(int drv);
	virtual bool WriteProtectedFloppyDisk(int drv);

	virtual bool OpenFloppyDisk(int drv, const _TCHAR* filename, int bank_num, uint32 flags, bool multiopen);
	virtual void CloseFloppyDisk(int drv);

	virtual bool OpenFloppyDiskSelectedVolume(int drv, int bank_num);

	virtual void GetMultiVolumeStr(int num, const _TCHAR *name, _TCHAR *str);

	virtual d88_file_t *GetD88File(int drv);
#endif

#ifdef USE_PRINTER
	// Printer

	virtual void PostEtSavePrinterMessage(int drv, const _TCHAR *file_path);
	virtual int  GetPrinterBufferSize(int drv);
	virtual void PostEtClearPrinterBufferMessage(int drv);
	virtual void PostEtPrintPrinterMessage(int drv);
	virtual void PostEtEnablePrinterDirectMessage(int drv);
	virtual bool IsEnablePrinterDirect(int drv);
#endif

	// Comm

	virtual void PostEtEnableCommServerMessage(int drv);
	virtual bool IsEnableCommServer(int drv);
	virtual void PostEtConnectCommMessage(int drv);
	virtual bool NowConnectingComm(int drv);

	// Status

	virtual void PostEtLoadStatusMessage(const _TCHAR *file_path);
	virtual void PostEtSaveStatusMessage(const _TCHAR *file_path, bool sys_pause);

	// Auto Key

	virtual void PostEtLoadAutoKeyMessage(const _TCHAR *file_path);
	virtual void PostEtStartAutoKeyMessage(void);
	virtual void PostEtStopAutoKeyMessage(void);
	virtual bool IsRunningAutoKey(void);

	virtual bool StartAutoKey(void);

	// Record key

	virtual void PostEtLoadRecKeyMessage(const _TCHAR *file_path);
	virtual void PostEtSaveRecKeyMessage(const _TCHAR *file_path, bool sys_pause);
	virtual void StopPlayRecKey(void);
	virtual void StopRecordRecKey(void);
	virtual bool NowPlayingRecKey(void);
	virtual bool NowRecordingRecKey(void);

	// Misc

	virtual void ToggleLedBox(void);
	virtual void ToggleInsideLedBox(void);
	virtual bool IsShownLedBox(void);
	virtual bool IsInsidedLedBox(void);
	virtual void ChangeLedBoxPosition(int num);
	virtual void ToggleMessageBoard(void);
	virtual bool IsShownMessageBoard(void);

	virtual void ToggleUseJoypad(void);
	virtual bool IsEnableJoypad(void);
	virtual void ToggleEnableLightpen(void);
	virtual bool IsEnableLightpen(void);
	virtual void ToggleUseMouse(void);
	virtual bool IsEnableMouse(void);

	virtual bool GetRecentFileStr(const _TCHAR *file, int num, _TCHAR *str, int trimlen);

	// Drop
	virtual bool OpenFileByExtention(const _TCHAR *file_path);
	virtual bool OpenDroppedFile(void *param);
	//@}

	/// @name utilities
	//@{
	static int CheckSupportedFile(const _TCHAR *file_path);
	static void GetLibVersionString(_TCHAR *str, int max_len = _MAX_PATH, const _TCHAR *sep_str = _T("\n"));
	//@}
};

#endif /* _GUI_BASE_H_ */
