/** @file rec_video.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2015.05.18 -

	@brief [ record video ]
*/

#include <Windows.h>
#include "rec_video.h"
#include "../emu.h"

#ifdef USE_REC_VIDEO_VFW
#include "vfw/vfw_rec_video.h"
#endif
#ifdef USE_REC_VIDEO_FFMPEG
#include "ffmpeg/ffm_rec_video.h"
#endif

REC_VIDEO::REC_VIDEO(EMU *new_emu)
{
	emu = new_emu;
	now_recording = false;
	rec_type = 0;
	rec_fps = 0;
	rec_count = 0;
	rec_denomi = 1;
	memset(rec_path, 0, sizeof(rec_path));

	dibRecVideo = new DibSection();

#ifdef USE_REC_VIDEO_VFW
	vfwvideo = new VFW_REC_VIDEO(new_emu, this);
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	ffmvideo = new FFM_REC_VIDEO(new_emu, this);
#endif
}

REC_VIDEO::~REC_VIDEO()
{
#ifdef USE_REC_VIDEO_FFMPEG
	delete ffmvideo;
#endif
#ifdef USE_REC_VIDEO_VFW
	delete vfwvideo;
#endif
	delete dibRecVideo;
}

bool REC_VIDEO::IsEnabled(int type)
{
	bool rc = false;
	switch(type) {
#ifdef USE_REC_VIDEO_VFW
	case RECORD_VIDEO_TYPE_VFW:
		rc = true;
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		rc = ffmvideo->IsEnabled();
		break;
#endif
	default:
		break;
	}
	return rc;
}

const _TCHAR **REC_VIDEO::GetCodecList(int type)
{
	const _TCHAR **list = NULL;
	switch(type) {
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		list = ffmvideo->GetCodecList();
		break;
#endif
	default:
		break;
	}
	return list;
}

const _TCHAR **REC_VIDEO::GetQualityList(int type)
{
	const _TCHAR **list = NULL;
	switch(type) {
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		list = ffmvideo->GetQualityList();
		break;
#endif
	default:
		break;
	}
	return list;
}

bool REC_VIDEO::Start(int type, int fps, VmRectWH *srcrect, DibSection *srcsurface, bool show_dialog)
{
	if (type > 0) {
		rec_type = type;
	}
	if(fps > 0) {
		rec_fps = fps;
		rec_count = 0;
		rec_denomi = 60 / fps;
	}

	if (rec_type <= 0 || rec_fps <= 0) {
		return false;
	}

	CreateFileName(rec_path);

#ifdef USE_REC_VIDEO_VFW
	if (rec_type == RECORD_VIDEO_TYPE_VFW) {
		if (srcrect != NULL) CreateSurface(srcrect->w, srcrect->h);
		now_recording = vfwvideo->Start(rec_path, rec_fps, srcrect, dibRecVideo, show_dialog);
	}
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	if (rec_type == RECORD_VIDEO_TYPE_FFMPEG) {
		// height set minus, so create the flipped surface
		if (srcrect != NULL) CreateSurface(srcrect->w, -srcrect->h);
		now_recording = ffmvideo->Start(rec_path, rec_fps, srcrect, dibRecVideo, show_dialog);
	}
#endif
	return now_recording;
}

void REC_VIDEO::Stop()
{
	if (!now_recording) return;
#ifdef USE_REC_VIDEO_VFW
	if (rec_type == RECORD_VIDEO_TYPE_VFW) {
		vfwvideo->Stop();
	}
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	if (rec_type == RECORD_VIDEO_TYPE_FFMPEG) {
		ffmvideo->Stop();
	}
#endif
	now_recording = false;
}

bool REC_VIDEO::Restart()
{
	if (!now_recording) return false;
#ifdef USE_REC_VIDEO_VFW
	if (rec_type == RECORD_VIDEO_TYPE_VFW) {
		now_recording = vfwvideo->Restart();
	}
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	if (rec_type == RECORD_VIDEO_TYPE_FFMPEG) {
		now_recording = ffmvideo->Restart();
	}
#endif
	return now_recording;
}

bool REC_VIDEO::Record(VmRectWH *srcrect, DibSection *srcsurface)
{
	if (!now_recording) return false;

	BitBlt(dibRecVideo->GetDC(), 0, 0, srcrect->w, srcrect->h, srcsurface->GetDC(), srcrect->x, srcrect->y, SRCCOPY);

#ifdef USE_REC_VIDEO_VFW
	if (rec_type == RECORD_VIDEO_TYPE_VFW) {
		now_recording = vfwvideo->Record(dibRecVideo);
	}
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	if (rec_type == RECORD_VIDEO_TYPE_FFMPEG) {
		now_recording = ffmvideo->Record(dibRecVideo);
	}
#endif
	return now_recording;
}

bool REC_VIDEO::IsRecordFrame()
{
	if (!now_recording) return false;

	bool rc = (rec_count == 0);
	rec_count = ((rec_count + 1) % rec_denomi);
	return rc;
}

bool REC_VIDEO::Capture(int type, VmRectWH *srcrect, DibSection *srcsurface)
{
	if (srcrect == NULL || srcsurface <= 0) {
		return false;
	}

	_TCHAR file_path[_MAX_PATH];
	CreateFileName(file_path);

	if (type == CAPTURE_SCREEN_TYPE_BMP) {
		// add ext
		_tcscat(file_path, _T(".bmp"));

		HDC hdc = GetDC(emu->get_window());
		DibSection dibBmp(hdc, srcrect->w, srcrect->h);
		ReleaseDC(emu->get_window(), hdc);
		if (!dibBmp.IsEnable()) {
			emu->out_log(LOG_ERROR,_T("Capture screen failed. Cannot create dib."));
			return false;
		}

		BitBlt(dibBmp.GetDC(), 0, 0, srcrect->w, srcrect->h, srcsurface->GetDC(), srcrect->x, srcrect->y, SRCCOPY);

		BITMAPFILEHEADER bmFileHeader = { (WORD)(TEXT('B') | TEXT('M') << 8) };
		bmFileHeader.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
		bmFileHeader.bfSize = bmFileHeader.bfOffBits + dibBmp.GetBufferSize();

		DWORD dwSize;
		HANDLE hFile = CreateFile(file_path, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (hFile == INVALID_HANDLE_VALUE) {
			emu->out_logf(LOG_ERROR,_T("CreateFile failed to open %s. code=%ld"),GetLastError());
			return false;
		}
		WriteFile(hFile, &bmFileHeader, sizeof(BITMAPFILEHEADER), &dwSize, NULL);
		WriteFile(hFile, dibBmp.GetHeader(), sizeof(BITMAPINFOHEADER), &dwSize, NULL);
		WriteFile(hFile, dibBmp.GetBuffer(), dibBmp.GetBufferSize(), &dwSize, NULL);
		CloseHandle(hFile);

	} else {
		emu->out_log(LOG_ERROR,_T("Capture screen failed. Unknown type."));
		return false;
	}
	emu->out_log(LOG_INFO,_("Screen was saved successfully."));
	return true;
}

void REC_VIDEO::CreateFileName(_TCHAR *file_path)
{
	int tim[8];
	const _TCHAR *app_path;

	app_path = _tcslen(config.snapshot_path) > 0 ? config.snapshot_path : emu->application_path();
	emu->get_timer(tim);
	_stprintf(file_path, _T("%s%04d-%02d-%02d_%02d-%02d-%02d"), app_path, tim[0], tim[1], tim[2], tim[4], tim[5], tim[6]);
}

void REC_VIDEO::CreateSurface(int w, int h)
{
	HDC hdc = GetDC(emu->get_window());
	dibRecVideo->CreateDib(hdc, w, h);
	ReleaseDC(emu->get_window(), hdc);
}
