/** @file display.cpp

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.06.08 -

	@brief [ display ]
*/

#include "display.h"
#include "../../fileio.h"
#include "../../config.h"

void DISPLAY::initialize()
{
	memset(font, 0, sizeof(font));
	font_rom_loaded = false;
	font_rom_loaded_at_first = false;

	// load rom images
	load_font_rom_file();

	emu->get_rgbformat(&Rmask, &Gmask, &Bmask, &Rshift, &Gshift, &Bshift);

	// create pc palette
	for(int i = 0; i < 8; i++) {
//		palette_pc[i] = RGB_COLOR((i & 2) ? 255 : 0, (i & 4) ? 255 : 0, (i & 1) ? 255 : 0);
		palette_pc[i] = emu->map_rgbcolor((i & 2) ? 255 : 0, (i & 4) ? 255 : 0, (i & 1) ? 255 : 0);
		emu->out_debugf(_T("palette_pc[%d]:%x"),i,palette_pc[i]);
	}

	memset(scrntop, 0, sizeof(scrntop));

	screen_left = 0;
	screen_top = 0;
	screen_right = SCREEN_WIDTH + screen_left;
	screen_bottom = SCREEN_HEIGHT + screen_top;
	scrnline1_offset = 0;

	mode_sel = 0;
	ig_enable = 0;
	ig_mask = true;

	start_addr = 0x4000;
	cursor_addr = 0;
	cblink = 0;
	disp_cursor = false;
	for(int i=0; i<32; i++) {
		cur_left[i] = 0;
		cur_right[i] = 0;
	}
	cur_top = 0;
	cur_bottom = 0;

	crtc_chr_clocks = 0;
	crtc_ppch = 16;

	interlace = 0;
	videomode = 0;
	disptmg_skew = 0;
	curdisp_skew = 0;

	bg_color = 0;
	hireso = 0;
	width = 1;
	width_sel = 0;

	top_raster = 0;
	bottom_raster = 0;
	vsync_top_raster = 0;
	vsync_bottom_raster = 0;
	raster_per_char = 16;
	now_raster = 0;
	lines_per_2frames = (LINES_PER_FRAME << 1);

	// synchronizable vertical range
	v_total_min[0] = (LINES_PER_FRAME - 24);
	v_total_max[0] = (LINES_PER_FRAME + 24);
	v_total_min[1] = (LINES_PER_FRAME - 24) / 2;
	v_total_max[1] = (LINES_PER_FRAME + 24) / 2;
	v_total_min[2] = (LINES_PER_FRAME - 24) * 2;
	v_total_max[2] = (LINES_PER_FRAME + 24) * 2;

	for(int i=0; i<VRAM_BUF_HEIGHT; i++) {
		left_col[i] = 0;
		right_col[i] = 0;
		hsync_left_col[i] = 0;
		hsync_right_col[i] = 0;
	}
	left_diff = 0;
	// synchronizable horizontal range
	h_total_min[0] = CHARS_PER_LINE - 3;	// w.40
	h_total_max[0] = CHARS_PER_LINE + 3;
	h_total_min[1] = CHARS_PER_LINE * 2 - 4;	// w.80
	h_total_max[1] = CHARS_PER_LINE * 2 + 4;

	sw = 0;

	Rloss = (64 << Rshift);
	Gloss = (32 << Gshift);
	Bloss = (96 << Bshift);

	for(int i=0; i<8; i++) {
		ppd[i]=(mode_sel & 0x80) ? 1 : 2;	// pixel per 1dot
		dof[i]=(0x80 >> i);
	}
	// pixel per 1char
	ppch = (mode_sel & 0x80) ? 8 : 16;

	//
	memset(crt_mon, 0, sizeof(crt_mon));
	crt_mon_col = 0;
	crt_mon_row = 0;
	crt_mon_addr = ADDR_VRAM_START;
	crt_mon_raster = 0;
	crt_mon_page = 0;

	crt_mon_cur_col = 0;
	crt_mon_cur_row = 0;

	skip_frame_count = 0;

	update_config();

	// register event (unused)
//	register_frame_event(this);
}

void DISPLAY::reset()
{
	load_font_rom_file();

	ig_enable = 0;
	ig_mask = true;
}

void DISPLAY::update_config()
{
	scanline = config.scan_line;
	afterimage = config.afterimage;
	keepimage = config.keepimage;

	// color filter for afterimage
	if (afterimage == 0) {
		memset(scrnline1, 0x00, sizeof(scrntype) * SCREEN_WIDTH);
		memset(scrnline2, 0xff, sizeof(scrnline2));
		scrnline1_offset = 0;
	} else if (afterimage == 1) {
		memset(scrnline2, 0xff, sizeof(scrnline2));
		scrnline1_offset = SCREEN_WIDTH;
	} else if (afterimage == 2) {
		memset(scrnline2, 0xbf, sizeof(scrnline2));
		scrnline1_offset = SCREEN_WIDTH;
	}

	if (keepimage > 0) {
		memset(scrntop, 0, sizeof(scrntop));
	}
}

void DISPLAY::load_font_rom_file()
{
	const _TCHAR *app_path, *rom_path[2];
	_TCHAR file_path[_MAX_PATH];

	rom_path[0] = config.rom_path;
	rom_path[1] = emu->application_path();

	FILEIO* fio = new FILEIO();
	for(int i=0; i<2; i++) {
		app_path = rom_path[i];

		if (!font_rom_loaded) {
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sFONT.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(font, sizeof(font), 1);
				fio->Fclose();
				font_rom_loaded = true;
				emu->out_log(LOG_INFO, (CT)_("FONT.ROM was loaded."));
			}
		}
	}
	delete fio;

	if (!font_rom_loaded) {
		if (!font_rom_loaded_at_first) {
			emu->out_log(LOG_WARN, (CT)_("FONT.ROM couldn't be loaded."));
		}
		memset(chrfont, 0, sizeof(chrfont));
	} else {
		// convert font data to useful array data
		for(int i = 0; i < 2; i++) {
			for(int c = 0; c < 256; c++) {
				for(int l = 0; l < 16; l++) {
					chrfont[i][c][l] = get_font_data(i, c, l);
				}
			}
		}
	}
	font_rom_loaded_at_first = true;
}

void DISPLAY::write_io8(uint32 addr, uint32 data)
{
	switch(addr & 0xffff) {
	case 0xffc6:
		// crtc addr register
		crtc_addr = (data & 0xff);
		break;
	case 0xffc7:
		// crtc data register
		switch(crtc_addr) {
			case 1:
				width = regs[1] ? regs[1] : 1;					// width
				break;
			case 8:
				interlace = (regs[8] & 1) ? 1 : 0;				// 0:noninterlace 1:interlace
				videomode = ((regs[8] & 3) == 3) ? 1 : 0;		// 1:videomode(interlace only)

				disptmg_skew = ((regs[8] & 0x30) >> 4);			// disptmg skew
				curdisp_skew = ((regs[8] & 0xc0) >> 6);			// curdisp skew
//				break;	// through
			case 9:
				// raster of 1char
				raster_per_char = (regs[9] & 0x1f) + videomode + 1;
				raster_per_char = raster_per_char << (1 - videomode);
				break;

		}
		break;
	case 0xffd0:
		// mode sel
		mode_sel = (data & 0xff);

		bg_color = mode_sel & 0x07;				// bg color
		hireso = (mode_sel & 0x40) ? 0 : 1;		// 0:normal 1:hireso
		width_sel = (mode_sel & 0x80) ? 1 : 0;	// 0:w.40  1:w.80

		for(int i=0; i<8; i++) {
			ppd[i]=width_sel ? 1 : 2;	// pixel per 1dot
			dof[i]=(0x80 >> i);
		}
		// pixel per 1char
		ppch = (mode_sel & 0x80) ? 8 : 16;

		crtc_chr_clocks = width_sel;
		crtc_ppch = ppch;

		break;
	case 0xffd6:
		// interlace sel
		interlace_sel = (data & 0x08) ? 1 : 0;	// 0:noninterlace char  1:interlace char
		break;
	case 0xffe9:
		// ig mode reg
		ig_enable = (data & 0x01);
		ig_mask &= ((data & 0x100) == 0);
		break;
	}
}

void DISPLAY::write_signal(int id, uint32 data, uint32 mask)
{
	switch(id) {
		case SIG_DISPLAY_LPSTB:
			{
			int width = regs[1];
			int dotx = (data >> 16) & 0xffff;
			int doty = (data & 0xffff);
			int x = 0;
			int y = 0;
			int offset = (width_sel ? 5 : 4);
			uint32 newaddr;

			if (dotx < left_col[doty >> 1]) { dotx = right_col[doty >> 1] - 1; doty-=16; }
			if (dotx >= right_col[doty >> 1]) { dotx = left_col[doty >> 1]; doty+=16; }
			if (doty < top_raster) doty = top_raster;
			if (doty >= bottom_raster) doty = bottom_raster - 1;

			x = (dotx - left_col[doty >> 1]) >> (width_sel ? 3 : 4);
			y = (doty - top_raster) >> 4;
			newaddr = ((regs[12] << 8)+regs[13]) + y * width + x + offset;
			regs[16] = (newaddr >> 8) & 0x3f;
			regs[17] = newaddr & 0xff;

//			emu->out_debugf("lp dx:%03d dy:%03d x:%02d y:%02d addr:%02x%02x"
//				,dotx,doty,x,y,regs[16],regs[17]);
			}
			break;

		case SIG_DISPLAY_VSYNC:
			if ((data & mask) == 0) {
				// synchronize display vertical raster
				for(int i=0; i<3; i++) {
					if ((v_total_min[i] <= now_raster && now_raster <= v_total_max[i])
					|| (v_total_min[i] <= *crtc_vt_total && *crtc_vt_total <= v_total_max[i])) {
						now_raster = 0;
						break;
					}
				}
				// cursor blink
				cblink = (cblink + 1) & 0x1f;

//				emu->out_debugf("vsync now:%d vt_total:%d vt_count:%d"
//				,now_raster,*vt_total,*vt_count);
			}
			break;
	}
}

void DISPLAY::update_cursor()
{
	int bp = regs[10] & 0x60;						// cursor blink speed

	uint8 cur_st_l;
	uint8 cur_ed_l;

	if(crt_mon_cur_col < 0 || crt_mon_cur_row < 0 || curdisp_skew == 3) {
		disp_cursor = false;
	} else {
		disp_cursor = (bp == 0 || (bp == 0x40 && (cblink & 8)) || (bp == 0x60 && (cblink & 0x10)));
	}

	if (disp_cursor) {
		if (videomode == 1) {
			cur_st_l = (regs[10] & 0x0f);
			cur_ed_l = (regs[11] & 0x0f);
		} else {
			cur_st_l = ((regs[10] & 0x0f) << 1);
			cur_ed_l = ((regs[11] & 0x0f) << 1) + 1;
		}

		// cursor draw area
		cur_top = top_raster + crt_mon_cur_row * 2;
		cur_bottom = cur_top + cur_ed_l;
		cur_top += cur_st_l;

		for(int i=0; i < raster_per_char; i++) {
			cur_left[i] = left_col[crt_mon_cur_row] + (crt_mon_cur_col + curdisp_skew + config.curdisp_skew - 2 - disptmg_skew + config.disptmg_skew) * crtc_ppch;
			cur_right[i] = cur_left[i] + crtc_ppch;
		}
	}
	crt_mon_cur_col = -1;
	crt_mon_cur_row = -1;
}

void DISPLAY::set_display_size(int left, int top, int right, int bottom)
{
	screen_left = left;
	screen_right = right;
	screen_top = top;
	screen_bottom = bottom;
}

void DISPLAY::draw_screen()
{
	if (emu->now_skip_frame()) {
		skip_frame_count++;
	} else {
		skip_frame_count = 0;
	}
	if ((afterimage == 0 && skip_frame_count >= 1)
	||  (afterimage > 0 && skip_frame_count >= 2)) {
		// clear cursor
		disp_cursor = false;
		crt_mon_cur_col = -1;
		crt_mon_cur_row = -1;

		return;
	}

	sw = 1 - sw;

	// update cursor
	update_cursor();

	// render screen
	draw_screen_sub();
}

uint8 DISPLAY::get_font_data(int interlace, int chr_code, int y)
{
	int code = 0;
	int offset = 0;
	if (interlace == 1) {
		// interlace mode char
//		if (videomode == 0) {
//			y = y >> 1;	//	double size
//		}
		if (chr_code < 0x80) {
			code = chr_code << 4;	// 16byte per 1char
			offset = (y & 0x0e);	// 0 0 2 2 4 4 6 6 8 8 a a c c e e
		} else {
			code = chr_code << 4;	// 16byte per 1char
			offset = (y & 0x0f);	// 0 1 2 3 4 5 6 7 8 9 a b c d e f
		}
	} else {
		// non interlace mode char
		if (chr_code < 0x80) {
			// ascii code
			code = chr_code << 4;	// 16byte per 1char
		} else {
			code = ((chr_code - 0x80) << 4) + 1;
		}
//		if (videomode == 0) {
//			offset = (y & 0x0e);		// 0 0 2 2 4 4 6 6 8 8 a a c c e e
//		} else {
			offset = ((y << 1) & 0x0e);	// 0 2 4 6 8 a c e 0 2 4 6 8 a c e
//		}
	}
	return font[code + offset];
}

#define COLOR_REVERSE(fg,bg) { \
	{ \
		uint8 c; \
		c = (bg); \
		(bg) = (fg); \
		(fg) = c; \
	} \
}

#define DISP_OUT_OF_DISPTMG { \
	disp_cursor_char = (disp_cursor_line && cur_left[l] <= x && x < cur_right[l]); \
\
	if (hsync_left_col[buf_y_h] <= x && x < hsync_right_col[buf_y_h]) { \
		c_f = 0; \
		c_b = 0; \
	} else { \
		c_f = cram & 0x07; \
		c_b = bg_color; \
	} \
	if (disp_cursor_char) COLOR_REVERSE(c_f, c_b); \
	scrn[x] = (palette_pc[c_b] & scrnline2[x]) | scrnline1p[x]; \
}

#define DISP_TEXT_CHAR_MAIN(code, c_f, c_b, scrn1, palette) { \
	bi = 0; \
	for(bt = 0; bt < 8; bt++) { \
		fg = (code & dof[bt]) ? c_f : c_b; \
		dtm = x+bi+ppd[bt]; \
		if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH; \
		for(dt = x+bi; dt < dtm; dt++) { \
			scrn1 = (palette); \
		} \
		bi += ppd[bt]; \
	} \
}

#define DISP_TEXT_CHAR { \
	p = chrfont[interlace_sel][code][(l >> (1 - videomode)) & 0xf]; \
\
	if (disp_cursor_char) COLOR_REVERSE(c_f, c_b); \
	DISP_TEXT_CHAR_MAIN(p, c_f, c_b, scrn[dt], (palette_pc[fg] & scrnline2[dt]) | scrnline1p[dt]) \
	if (ki) { \
		DISP_TEXT_CHAR_MAIN(p, ck_f, ck_b, scrntop[y][dt], palette_pc[fg]) \
	} \
}

#define DISP_IG_PATTERN_MAIN(c_f, c_b, scrn1, palette) { \
	bi = 0; \
	for(bt = 0; bt < 8; bt++) { \
		p = ((i_g & dof[bt]) ? 4 : 0) | ((i_r & dof[bt]) ? 2 : 0) | ((i_b & dof[bt]) ? 1 : 0); \
		if (p == 0) p = c_b; \
		dtm = x+bi+ppd[bt]; \
		if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH; \
		for(dt = x+bi; dt < dtm; dt++) { \
			scrn1 = (palette); \
		} \
		bi += ppd[bt]; \
	} \
}

#define DISP_IG_PATTERN { \
	ll = l >> int_vid; \
	i_b = ig_ram[code*8+ll]; \
	i_r = ig_ram[code*8+ll+IGRAM_SIZE]; \
	i_g = ig_ram[code*8+ll+IGRAM_SIZE+IGRAM_SIZE]; \
\
	DISP_IG_PATTERN_MAIN(c_f, c_b, scrn[dt], (palette_pc[p] & scrnline2[dt]) | scrnline1p[dt]) \
	if (ki) { \
		DISP_IG_PATTERN_MAIN(0, 0, scrntop[y][dt], palette_pc[p]) \
	} \
}

#define DISP_WHITE_SPACE_MAIN(scrn1, palette) { \
	for(dt = x; dt < dtm; dt++) { \
		scrn1 = (palette); \
	} \
}

#define DISP_WHITE_SPACE { \
	p = 0xff; \
	c_f = 7; \
	c_b = bg_color; \
	ck_f = 7; \
	ck_b = 0; \
\
	if (buf_x == 0 && buf_y_h == 0) { \
		bi = 0; \
		int btt = (1 << (rand() % 4)) | 0x20; \
		for(bt = 0; bt < 8; bt++) { \
			p = (btt & (1 << (7-bt))) != 0 ? 3 : 7; \
			dtm = x+bi+ppd[bt]; \
			if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH; \
			for(dt = x+bi; dt < dtm; dt++) { \
				scrn[dt] = (palette_pc[p] & scrnline2[dt]) | scrnline1p[dt]; \
			} \
			bi += ppd[bt]; \
		} \
	} else { \
		dtm = x+ppch; \
		if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH; \
		DISP_WHITE_SPACE_MAIN(scrn[dt], (palette_pc[c_f] & scrnline2[dt]) | scrnline1p[dt]) \
	} \
	if (ki) { \
		DISP_WHITE_SPACE_MAIN(scrntop[y][dt], palette_pc[ck_f]) \
	} \
}

#define	DISP_GRAPHIC_HIRESO_MAIN(code, c_f, c_b, scrn1, palette) { \
	bi = 0; \
	for(bt = 0; bt < 8; bt++) { \
		fg = (code & dof[bt]) ? c_f : c_b; \
		dtm = x+bi+ppd[bt]; \
		if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH; \
		for(dt = x+bi; dt < dtm; dt++) { \
			scrn1 = (palette); \
		} \
		bi += ppd[bt]; \
	} \
}

#define	DISP_GRAPHIC_HIRESO { \
	DISP_GRAPHIC_HIRESO_MAIN(code, c_f, c_b, scrn[dt], (palette_pc[fg] & scrnline2[dt]) | scrnline1p[dt]) \
	if (ki) { \
		DISP_GRAPHIC_HIRESO_MAIN(code, ck_f, ck_b, scrntop[y][dt], palette_pc[fg]) \
	} \
}

#define DISP_GRAPHIC_NORMAL_MAIN(code, c_f, c_b, scrn1, palette) { \
	bi = 0; \
	for (bt = 0; bt < 8; bt += 4) { \
		fg = (code & ((0x10 >> bt) << (l >> 2))) ? c_f : c_b; \
		dtm = x+bi+ppd[bt]+ppd[bt+1]+ppd[bt+2]+ppd[bt+3]; \
		if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH; \
		for(dt = x+bi; dt < dtm; dt++) { \
			scrn1 = (palette); \
		} \
		bi += ppd[bt] + ppd[bt + 1] + ppd[bt + 2] + ppd[bt + 3]; \
	} \
}

#define DISP_GRAPHIC_NORMAL { \
	DISP_GRAPHIC_NORMAL_MAIN(code, c_f, c_b, scrn[dt], (palette_pc[fg] & scrnline2[dt]) | scrnline1p[dt]) \
	if (ki) { \
		DISP_GRAPHIC_NORMAL_MAIN(code, ck_f, ck_b, scrntop[y][dt], palette_pc[fg]) \
	} \
}

void DISPLAY::draw_screen_sub()
{
	static uint8 p;
	static uint8 code;
	static uint8 c_f;
	static uint8 c_b;
	static uint8 ck_f;
	static uint8 ck_b;
	static uint8 cram = 0;
	static uint8 fg;
	static uint8 i_r,i_g,i_b;
	static int bt;
	static int bi;
	static int dt, dtm;
	static int x,x0,x1;
	static int y,y0[3],y1[3],ys,yd;
	static int l,ll;
	static int int_vid;
	static bool disp_cursor_line;
	static bool disp_cursor_char;
	static bool ki;
//	bool skip_line;
	static int buf_x = 0;
	static int buf_y;
	static int buf_y_h;
	static int i,ii[2];

	static int scrn_offset;
	static scrntype *scrn0;
	static scrntype *scrn;
	static scrntype dot;
	static uint32 r,g,b;

	static scrntype *scrnline1p;

	disp_cursor_line = false;
	disp_cursor_char = false;
	scrn_offset = emu->screen_buffer_offset();
	scrn0 = emu->screen_buffer(0);
	dot = 0;
	buf_x = 0;

	// interace or videomode
	int_vid = ((interlace_sel) + (1 - videomode));

	// color filter for afterimage
	if (afterimage == 1) {
		scrn = scrn0 + scrn_offset * screen_top;
		for(y = screen_top; y < screen_bottom; y++) {
			for(x = screen_left; x < screen_right; x++) {
				dot =scrn[x];
				r = (dot & Rmask);
				g = (dot & Gmask);
				b = (dot & Bmask);

				if (r > Rloss) r = r - Rloss;
				else r = 0;
				if (g > Gloss) g = g - Gloss;
				else g = 0;
				if (b > Bloss) b = b - Bloss;
				else b = 0;

				dot = r | g | b;

				scrnline1[y][x] = dot;
			}
			scrn += scrn_offset;
		}
	} else if (afterimage == 2) {
		scrn = scrn0 + scrn_offset * screen_top;
		for(y = screen_top; y < screen_bottom; y++) {
			for(x = screen_left; x < screen_right; x++) {
				// 0xff or 0xdf -> 0xe0 -> 0x00
//				dot = scrn[x] & 0x10101010;
//				dot = (dot << 3) | (dot << 2) | (dot << 1);
//				scrnline2[x] = 0xdfdfdfdf;
				// 0xff or 0xbf -> 0xc0 -> 0x00
				dot = scrn[x] & 0x20202020;
				dot = (dot << 2) | (dot << 1);

				scrnline1[y][x] = dot;
			}
			scrn += scrn_offset;
		}
	}

	// scanline
	ys = (scanline ? 2 : 1);	// skip step
	yd = (scanline && (interlace == 1 && sw == (screen_top & 1)) ? 1 : 0);	// odd line ?

	// set y axis position from the kind of drawing line
	if (disptmg_skew == 3) {
		y0[0] = (screen_top | yd); y1[0] = screen_bottom;	// no drawing
		y0[1] = 0; y1[1] = 0;
		y0[2] = 0; y1[2] = 0;
		ii[0] = 1; ii[1] = 1;
	}
	else if (top_raster <  bottom_raster) {
		y0[0] = (screen_top | yd); y1[0] = (screen_bottom < top_raster ? screen_bottom : top_raster);	// top padding
		y0[1] = (bottom_raster | yd); y1[1] = screen_bottom;	// bottom padding
		y0[2] = ((top_raster > screen_top ? top_raster : screen_top) | yd); y1[2] = (screen_bottom < bottom_raster ? screen_bottom : bottom_raster);
		ii[0] = 2; ii[1] = 3;
	}
	else if (top_raster >= bottom_raster) {
		y0[0] = ((bottom_raster > screen_top ? bottom_raster : screen_top) | yd); y1[0] = (screen_bottom < top_raster ? screen_bottom : top_raster);
		y0[1] = (screen_top | yd); y1[1] = (screen_bottom < bottom_raster ? screen_bottom : bottom_raster);	// top drawing
		y0[2] = (top_raster | yd); y1[2] = screen_bottom;	// bottom drawing
		ii[0] = 1; ii[1] = 3;
	}
	else {
		y0[0] = (screen_top | yd); y1[1] = screen_bottom;
		y0[1] = 0; y1[1] = 0;
		y0[2] = 0; y1[2] = 0;
		ii[0] = 0; ii[1] = 1;
	}

//emu->out_logf(LOG_DEBUG,_T("ii0:%d ii1:%d"),ii[0],ii[1]);
//for(i = 0; i < 3; i++) {
//	emu->out_logf(LOG_DEBUG,_T(" y:%d y0:%d y1:%d"),i,y0[i],y1[i]);
//}

	// out of disptmg area (top or bottom padding area)
	for(i = 0; i < ii[0]; i++) {
		buf_y = - top_raster + (VRAM_BUF_HEIGHT << 1) + y0[i];
		scrn = scrn0 + scrn_offset * y0[i];
		scrnline1p = scrnline1[0] + scrnline1_offset * y0[i];

		for(y = y0[i]; y < y1[i]; ) {
			buf_y = buf_y % (VRAM_BUF_HEIGHT << 1);
			buf_y_h = buf_y >> 1;

			if (vsync_top_raster <= y && y < vsync_bottom_raster) {
				for(x = screen_left; x < screen_right; x++) {
					scrn[x] = (palette_pc[0] & scrnline2[x]) | scrnline1p[x];
				}
			} else {
				for(x = screen_left; x < screen_right; x++) {
					if (hsync_left_col[buf_y_h] <= x && x < hsync_right_col[buf_y_h]) {
						scrn[x] = (palette_pc[0] & scrnline2[x]) | scrnline1p[x];
					} else {
						scrn[x] = (palette_pc[bg_color] & scrnline2[x]) | scrnline1p[x];
					}
				}
			}
			y+=ys;
			buf_y+=ys;
			scrn += scrn_offset * ys;
			scrnline1p += scrnline1_offset * ys;
		}
	}

	// disptmp area
	for(i = ii[0]; i < ii[1]; i++) {
		buf_y = - top_raster + (VRAM_BUF_HEIGHT << 1) + y0[i];
		scrn = scrn0 + scrn_offset * y0[i];
		l = (top_raster <= bottom_raster ? 0 : raster_per_char - ((bottom_raster - 1) % raster_per_char) - 1) | yd;
		scrnline1p = scrnline1[0] + scrnline1_offset * y0[i];

		for(y = y0[i]; y < y1[i]; ) {
			buf_y = buf_y % (VRAM_BUF_HEIGHT << 1);
			buf_y_h = buf_y >> 1;

			// cursor line ?
			disp_cursor_line = (disp_cursor && cur_top <= y && y <= cur_bottom);

			// out of disptmg area (left padding area)
			x1 = (screen_right < left_col[buf_y_h] ? screen_right : left_col[buf_y_h]);
			for(x = screen_left; x < x1; ) {
				DISP_OUT_OF_DISPTMG;
				x++;
			}
			// out of disptmg area (right padding area)
			x1 = (screen_right < right_col[buf_y_h] ? screen_right : right_col[buf_y_h]);
			for(x = x1; x < screen_right; ) {
				DISP_OUT_OF_DISPTMG;
				x++;
			}

			// disptmg area
			buf_x = left_col[buf_y_h] < 0 ? - left_col[buf_y_h] / ppch : 0 ;
			x0 = (screen_left > left_col[buf_y_h] ? screen_left : left_col[buf_y_h]);
			x1 = (screen_right < right_col[buf_y_h] ? screen_right : right_col[buf_y_h]);
			for(x = x0; x < x1; ) {
				disp_cursor_char = (disp_cursor_line && cur_left[l] <= x && x < cur_right[l]);

				code = crt_mon[buf_y_h][buf_x].text;
				cram = crt_mon[buf_y_h][buf_x].attr;
				ki = cram & 0x80 ? true : false;
				c_f = cram & 0x07;
				c_b = bg_color;
				ck_f = c_f;
				ck_b = 0;
				// color reverse
				if ((cram & 0x28) == 0x08) {
					COLOR_REVERSE(c_f, c_b);
					if (ki) {
						COLOR_REVERSE(ck_f, ck_b);
					}
				}

				switch(cram & 0x30) {
				case 0x30:
				case 0x20:
					// ig
					if (!ig_enable) {
						// disp ig pattern
						DISP_IG_PATTERN;
					} else {
						// disp white square
						DISP_WHITE_SPACE;
					}
					break;
				case 0x10:
					// graphic
					p = code;

					if (disp_cursor_char) COLOR_REVERSE(c_f, c_b);
					if (cram & 0x40) {
						// hireso graphic
						DISP_GRAPHIC_HIRESO;
					} else {
						// normal graphic
						DISP_GRAPHIC_NORMAL;
					}
					break;
				default:
					// character
					DISP_TEXT_CHAR;
					break;
				}
				x += ppch;
				buf_x++;
			}

			y+=ys;
			buf_y+=ys;
			scrn += scrn_offset * ys;
			l = ((l + ys) % raster_per_char);
			scrnline1p += scrnline1_offset * ys;
		}
	}

	// clear non drawing line
	if (scanline) {
		yd = 1 - yd;
		scrn = scrn0 + scrn_offset * (screen_top | yd);
		scrnline1p = scrnline1[0] + scrnline1_offset * (screen_top | yd);
		for(y = (screen_top | yd); y < screen_bottom; ) {
			for(x = screen_left; x < screen_right; x++) {
				scrn[x] = (palette_pc[0] & scrnline2[x]) | scrnline1p[x];
			}
			y+=ys;
			scrn += scrn_offset * ys;
			scrnline1p += scrnline1_offset * ys;
		}
	}

	// overlap the top most screen if need
	if (keepimage > 0) {
		scrn = emu->screen_buffer(screen_top);
		for(y = screen_top; y < screen_bottom; y++) {
			for(x = screen_left; x < screen_right; x++) {
				if (scrntop[y][x] != palette_pc[0]) scrn[x] = scrntop[y][x];
			}
			scrn += scrn_offset;
		}
	}
}

void DISPLAY::update_vram(int v, int clock)
{
	if (*crtc_vt_count != crt_mon_row) {
		if (*crtc_vt_count == 0) {
			// disptmg on

			// calculate start addr from crtc
			start_addr = (((regs[12] << 8) | regs[13]) + disptmg_skew - config.disptmg_skew) & 0x3fff;
			if (start_addr < ADDR_VRAM_START) {
				start_addr += (ADDR_VRAM_END - ADDR_VRAM_START);
			}

			// pixel of height
			top_raster = now_raster * 2;
			top_raster += (SCREEN_HEIGHT - 488) / 2;
			bottom_raster = top_raster + (regs[6] * raster_per_char);

			vsync_top_raster = top_raster + (regs[7] * raster_per_char);
			vsync_bottom_raster = vsync_top_raster + (((regs[3] & 0xf0) ? (regs[3] >> 4) : 16) << 1);

//			top_raster = (top_raster + lines_per_2frames) % lines_per_2frames;
//			bottom_raster = (bottom_raster + lines_per_2frames) % lines_per_2frames;

//			vsync_top_raster = (vsync_top_raster + lines_per_2frames) % lines_per_2frames;
//			vsync_bottom_raster = (vsync_bottom_raster + lines_per_2frames) % lines_per_2frames;

			// synchronize horizontal total
			if (h_total_min[crtc_chr_clocks] <= regs[0] && regs[0] <= h_total_max[crtc_chr_clocks]) {
				left_diff = 0;
			}
			// top most page ?
			crt_mon_page = ((keepimage == 1 && start_addr < 0x0800) || (keepimage == 2 && start_addr >= 0x0800)) ? 0x80 : 0;
		}

		crt_mon_row = *crtc_vt_count;
		crt_mon_col = 0;
		crt_mon_raster = (crt_mon_row % (raster_per_char >> 1));
		crt_mon_addr = start_addr + (crt_mon_row / (raster_per_char >> 1)) * width;
		crt_mon_addr += hireso * (crt_mon_raster & 0x7) * (0x400 << width_sel);

		// cursor
		cursor_addr = (((regs[14] << 8) | regs[15]) & 0x3fff) + disptmg_skew - config.disptmg_skew;
		if (cursor_addr < 0x400) cursor_addr += 0x4000;

		// pixel of width
		if (crt_mon_row < VRAM_BUF_HEIGHT) {
			left_col[crt_mon_row] = regs[0]-regs[2]-(regs[3] & 0x0f)-(crtc_chr_clocks ? 14 : 5) + disptmg_skew - config.disptmg_skew;
			left_col[crt_mon_row] *= crtc_ppch;
			left_col[crt_mon_row] += ((SCREEN_WIDTH - 768) / 2);
			left_col[crt_mon_row] += left_diff;

			right_col[crt_mon_row] = left_col[crt_mon_row] + crtc_ppch * regs[1];

			hsync_left_col[crt_mon_row] = left_col[crt_mon_row] + regs[2] * crtc_ppch;
			hsync_right_col[crt_mon_row] = hsync_left_col[crt_mon_row] + (regs[3] & 0x0f) * crtc_ppch;
		}
		// diff horizontal total display and crtc
		if (regs[0] < h_total_min[crtc_chr_clocks] || h_total_max[crtc_chr_clocks] < regs[0]) {
			left_diff += (regs[0] + 1 - (1 << (crtc_chr_clocks + 6))) * crtc_ppch;

			if (left_diff >= SCREEN_WIDTH) left_diff = -SCREEN_WIDTH + 1;
			else if (left_diff <= -SCREEN_WIDTH) left_diff = SCREEN_WIDTH - 1;
		}

		now_raster++;
		if (now_raster >= lines_per_2frames) now_raster = 0;
	}

	// copy vram data to buffer
	if (crt_mon_row < VRAM_BUF_HEIGHT && *crtc_vt_count < *crtc_vt_disp) {
		for(int i=0; i < (clock << width_sel); i++) {
			if (crt_mon_addr >= ADDR_VRAM_END) {
				crt_mon_addr -= (ADDR_VRAM_END - ADDR_VRAM_START);
			}
			if (crt_mon_col < width) {
				crt_mon[crt_mon_row][crt_mon_col].text = vram[crt_mon_addr];
				// bit 0-5: color ram   bit 6: hireso  bit 7:top_most
				crt_mon[crt_mon_row][crt_mon_col].attr = (color_ram[crt_mon_addr - ADDR_VRAM_START] | (hireso << 6) | crt_mon_page) & (ig_mask ? 0xdf : 0xff);
				// cursor
				if (crt_mon_addr == cursor_addr && crt_mon_raster == 0) {
					crt_mon_cur_col = crt_mon_col;
					crt_mon_cur_row = crt_mon_row;
				}
			} else if (crt_mon_col < VRAM_BUF_WIDTH) {
				crt_mon[crt_mon_row][crt_mon_col].text = 0;
				crt_mon[crt_mon_row][crt_mon_col].attr = 0;
			}
			crt_mon_col++;
			crt_mon_addr++;
		}
	}
}

// ----------------------------------------------------------------------------
// event handler
// ----------------------------------------------------------------------------

void DISPLAY::event_frame()
{
//	cblink = (cblink + (1 << config.cpu_power)) & 0x1f;
//	sw = 1 - sw;
}

// ----------------------------------------------------------------------------

void DISPLAY::save_state(FILEIO *fio)
{
	struct vm_state_st vm_state;

	//
	vm_state_ident.version = 2;
	vm_state_ident.size = sizeof(vm_state_ident) + sizeof(vm_state);

	// copy values
	memset(&vm_state, 0, sizeof(vm_state));
	vm_state.crtc_addr = crtc_addr;
	vm_state.ig_enable = ig_enable;
	vm_state.ig_mask = ig_mask ? 1 : 0;

	vm_state.chr_clocks = (crtc_chr_clocks & 0xff);
	fio->Fwrite(&vm_state_ident, sizeof(vm_state_ident), 1);
	fio->Fwrite(&vm_state, sizeof(vm_state), 1);
}

void DISPLAY::load_state(FILEIO *fio)
{
	vm_state_ident_t vm_state_i;
	struct vm_state_st vm_state;

	READ_STATE_CHUNK(fio, vm_state_i, vm_state);

	// require CRTC register
	for(crtc_addr = 0; crtc_addr < 16; crtc_addr++) {
		write_io8(0xffc7, regs[crtc_addr]);
	}

	crtc_addr = vm_state.crtc_addr;
	ig_enable = vm_state.ig_enable;
	ig_mask = vm_state.ig_mask ? true : false;

	if (vm_state_i.version >= 2) {
		crtc_chr_clocks = vm_state.chr_clocks;
		crtc_ppch = (crtc_chr_clocks ? 8 : 16);
	}

	// clear
	crt_mon_row = 9999;
	crt_mon_col = 0;

}

